@@ -1,6 +1,5 @@
 /*
  * Copyright (c) 2015, 2016, 2017, 2019 Nicira, Inc.
- * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -71,36 +70,6 @@ union ct_addr {
     struct in6_addr ipv6;
 };
 
-struct ct_endpoint {
-    union ct_addr addr;
-    union {
-        ovs_be16 port;
-        struct {
-            ovs_be16 icmp_id;
-            uint8_t icmp_type;
-            uint8_t icmp_code;
-        };
-    };
-};
-
-/* Verify that there is no padding in struct ct_endpoint, to facilitate
- * hashing in ct_endpoint_hash_add(). */
-BUILD_ASSERT_DECL(sizeof(struct ct_endpoint) == sizeof(union ct_addr) + 4);
-
-/* Changes to this structure need to be reflected in conn_key_hash()
- * and conn_key_cmp(). */
-struct conn_key {
-    struct ct_endpoint src;
-    struct ct_endpoint dst;
-
-    ovs_be16 dl_type;
-    uint16_t zone;
-    uint8_t nw_proto;
-};
-
-/* Verify that nw_proto stays uint8_t as it's used to index into l4_protos[] */
-BUILD_ASSERT_DECL(MEMBER_SIZEOF(struct conn_key, nw_proto) == sizeof(uint8_t));
-
 enum nat_action_e {
     NAT_ACTION_SRC = 1 << 0,
     NAT_ACTION_SRC_PORT = 1 << 1,
@@ -116,28 +85,7 @@ struct nat_action_info_t {
     uint16_t nat_action;
 };
 
-struct conn_lookup_ctx {
-    struct conn_key key;
-    struct conn *conn;
-    uint32_t hash;
-    bool reply;
-    bool icmp_related;
-    bool valid;
-};
-
-enum ct_direction {
-    CT_DIR_INIT,
-    CT_DIR_REP,
-    CT_DIR_NUM,
-};
-
-struct ct_match {
-    odp_port_t odp_port;
-    odp_port_t orig_in_port;
-    struct conn_key key;
-};
-
-struct conntrack *conntrack_init(void *dp);
+struct conntrack *conntrack_init(void);
 void conntrack_destroy(struct conntrack *);
 
 int conntrack_execute(struct conntrack *ct, struct dp_packet_batch *pkt_batch,
@@ -152,7 +100,7 @@ void conntrack_clear(struct dp_packet *packet);
 struct conntrack_dump {
     struct conntrack *ct;
     unsigned bucket;
-    struct cmap_position cm_pos;
+    struct cmap_cursor cursor;
     bool filter_zone;
     uint16_t zone;
 };
@@ -160,12 +108,12 @@ struct conntrack_dump {
 struct conntrack_zone_limit {
     int32_t zone;
     uint32_t limit;
-    atomic_count count;
+    uint32_t count;
     uint32_t zone_limit_seq; /* Used to disambiguate zone limit counts. */
 };
 
 struct timeout_policy {
-    struct cmap_node node;
+    struct hmap_node node;
     struct ct_dpif_timeout_policy policy;
 };
 
@@ -190,7 +138,6 @@ int conntrack_flush_tuple(struct conntrack *, const struct ct_dpif_tuple *,
 int conntrack_set_maxconns(struct conntrack *ct, uint32_t maxconns);
 int conntrack_get_maxconns(struct conntrack *ct, uint32_t *maxconns);
 int conntrack_get_nconns(struct conntrack *ct, uint32_t *nconns);
-int conntrack_get_stats(struct conntrack *ct, struct ct_dpif_stats *stats);
 int conntrack_set_tcp_seq_chk(struct conntrack *ct, bool enabled);
 bool conntrack_get_tcp_seq_chk(struct conntrack *ct);
 struct ipf *conntrack_ipf_ctx(struct conntrack *ct);
@@ -198,8 +145,5 @@ struct conntrack_zone_limit zone_limit_get(struct conntrack *ct,
                                            int32_t zone);
 int zone_limit_update(struct conntrack *ct, int32_t zone, uint32_t limit);
 int zone_limit_delete(struct conntrack *ct, uint16_t zone);
-
-bool conn_key_extract(struct conntrack *, struct dp_packet *, ovs_be16 dl_type,
-                      struct conn_lookup_ctx *, uint16_t zone);
-
+
 #endif /* conntrack.h */
