@@ -1,6 +1,5 @@
 /*
  * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2016, 2017 Nicira, Inc.
- * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -441,8 +440,6 @@ netdev_open(const char *name, const char *type, struct netdev **netdevp)
 
                 ovs_list_init(&netdev->saved_flags_list);
 
-                ovs_refcount_init(&netdev->refcount);
-
                 error = rc->class->construct(netdev);
                 if (!error) {
                     netdev_change_seq_changed(netdev);
@@ -462,11 +459,10 @@ netdev_open(const char *name, const char *type, struct netdev **netdevp)
                       name, type);
             error = EAFNOSUPPORT;
         }
-    } else {
-        ovs_refcount_ref(&netdev->refcount);
     }
 
     if (!error) {
+        netdev->ref_cnt++;
         *netdevp = netdev;
     } else {
         *netdevp = NULL;
@@ -480,11 +476,15 @@ netdev_open(const char *name, const char *type, struct netdev **netdevp)
  * 'netdev_' is null. */
 struct netdev *
 netdev_ref(const struct netdev *netdev_)
+    OVS_EXCLUDED(netdev_mutex)
 {
     struct netdev *netdev = CONST_CAST(struct netdev *, netdev_);
 
     if (netdev) {
-        ovs_refcount_ref(&netdev->refcount);
+        ovs_mutex_lock(&netdev_mutex);
+        ovs_assert(netdev->ref_cnt > 0);
+        netdev->ref_cnt++;
+        ovs_mutex_unlock(&netdev_mutex);
     }
     return netdev;
 }
@@ -560,17 +560,6 @@ netdev_get_tunnel_config(const struct netdev *netdev)
     }
 }
 
-int
-netdev_derive_tunnel_config(const struct netdev *upper_netdev,
-                            struct netdev *lower_netdev)
-{
-    if (upper_netdev->netdev_class->derive_tunnel_config) {
-        return upper_netdev->netdev_class->derive_tunnel_config(upper_netdev,
-                                                                lower_netdev);
-    }
-    return EOPNOTSUPP;
-}
-
 /* Returns the id of the numa node the 'netdev' is on.  If the function
  * is not implemented, returns NETDEV_NUMA_UNSPEC. */
 int
@@ -583,53 +572,30 @@ netdev_get_numa_id(const struct netdev *netdev)
     }
 }
 
-bool
-netdev_is_configured(const struct netdev *netdev)
-{
-    /* If no reconfigure callback is defined for this
-     * netdev, it is already configured.
-     */
-    return netdev->reconfigure_status == 0 ||
-           netdev->reconfigure_status == EOPNOTSUPP;
-}
-
-static void
-netdev_destroy(struct netdev *dev)
-    OVS_REQUIRES(netdev_mutex)
-{
-    const struct netdev_class *class = dev->netdev_class;
-    struct netdev_registered_class *rc;
-
-    netdev_uninit_flow_api(dev);
-
-    dev->netdev_class->destruct(dev);
-
-    if (dev->node) {
-        shash_delete(&netdev_shash, dev->node);
-    }
-    free(dev->name);
-    seq_destroy(dev->reconfigure_seq);
-    dev->netdev_class->dealloc(dev);
-
-    rc = netdev_lookup_class(class->type);
-    ovs_refcount_unref(&rc->refcnt);
-}
-
-static void
-netdev_unref_protected(struct netdev *dev)
-    OVS_REQUIRES(netdev_mutex)
-{
-    if (ovs_refcount_unref(&dev->refcount) == 1) {
-        netdev_destroy(dev);
-    }
-}
-
 static void
 netdev_unref(struct netdev *dev)
+    OVS_RELEASES(netdev_mutex)
 {
-    if (ovs_refcount_unref(&dev->refcount) == 1) {
-        ovs_mutex_lock(&netdev_mutex);
-        netdev_destroy(dev);
+    ovs_assert(dev->ref_cnt);
+    if (!--dev->ref_cnt) {
+        const struct netdev_class *class = dev->netdev_class;
+        struct netdev_registered_class *rc;
+
+        netdev_uninit_flow_api(dev);
+
+        dev->netdev_class->destruct(dev);
+
+        if (dev->node) {
+            shash_delete(&netdev_shash, dev->node);
+        }
+        free(dev->name);
+        seq_destroy(dev->reconfigure_seq);
+        dev->netdev_class->dealloc(dev);
+        ovs_mutex_unlock(&netdev_mutex);
+
+        rc = netdev_lookup_class(class->type);
+        ovs_refcount_unref(&rc->refcnt);
+    } else {
         ovs_mutex_unlock(&netdev_mutex);
     }
 }
@@ -640,6 +606,7 @@ netdev_close(struct netdev *netdev)
     OVS_EXCLUDED(netdev_mutex)
 {
     if (netdev) {
+        ovs_mutex_lock(&netdev_mutex);
         netdev_unref(netdev);
     }
 }
@@ -662,8 +629,7 @@ netdev_remove(struct netdev *netdev)
             netdev->node = NULL;
             netdev_change_seq_changed(netdev);
         }
-        netdev_unref_protected(netdev);
-        ovs_mutex_unlock(&netdev_mutex);
+        netdev_unref(netdev);
     }
 }
 
@@ -950,7 +916,7 @@ netdev_pop_header(struct netdev *netdev, struct dp_packet_batch *batch)
     size_t i, size = dp_packet_batch_size(batch);
 
     DP_PACKET_BATCH_REFILL_FOR_EACH (i, size, packet, batch) {
-        packet = netdev->netdev_class->pop_header(packet, false);
+        packet = netdev->netdev_class->pop_header(packet);
         if (packet) {
             /* Reset the offload flags if present, to avoid wrong
              * interpretation in the further packet processing when
@@ -1009,7 +975,7 @@ netdev_push_header(const struct netdev *netdev,
                          netdev_get_name(netdev));
         } else {
             netdev->netdev_class->push_header(netdev, packet, data);
-            pkt_metadata_reset(&packet->md, data->out_port);
+            pkt_metadata_init(&packet->md, data->out_port);
             dp_packet_batch_refill(batch, packet, i);
         }
     }
@@ -1472,7 +1438,7 @@ do_update_flags(struct netdev *netdev, enum netdev_flags off,
             sf->saved_flags = changed_flags;
             sf->saved_values = changed_flags & new_flags;
 
-            netdev_ref(netdev);
+            netdev->ref_cnt++;
             ovs_mutex_unlock(&netdev_mutex);
         }
     }
@@ -1551,8 +1517,7 @@ netdev_restore_flags(struct netdev_saved_flags *sf)
         ovs_mutex_lock(&netdev_mutex);
         ovs_list_remove(&sf->node);
         free(sf);
-        netdev_unref_protected(netdev);
-        ovs_mutex_unlock(&netdev_mutex);
+        netdev_unref(netdev);
     }
 }
 
@@ -1677,24 +1642,6 @@ netdev_set_policing(struct netdev *netdev, uint32_t kbits_rate,
             : EOPNOTSUPP);
 }
 
-
-/* Read the input rate limiting (policing) policy, if possible. */
-int
-netdev_get_policing(struct netdev *netdev,
-                    uint32_t *kbits_rate, uint32_t *kbits_burst,
-                    uint32_t *kpkts_rate, uint32_t *kpkts_burst)
-{
-    *kbits_rate = 0;
-    *kbits_burst = 0;
-    *kpkts_rate = 0;
-    *kpkts_burst = 0;
-
-    return (netdev->netdev_class->get_policing
-            ? netdev->netdev_class->get_policing(netdev,
-                    kbits_rate, kbits_burst, kpkts_rate, kpkts_burst)
-            : EOPNOTSUPP);
-}
-
 /* Adds to 'types' all of the forms of QoS supported by 'netdev', or leaves it
  * empty if 'netdev' does not support QoS.  Any names added to 'types' should
  * be documented as valid for the "type" column in the "QoS" table in
@@ -2074,7 +2021,7 @@ netdev_from_name(const char *name)
     ovs_mutex_lock(&netdev_mutex);
     netdev = shash_find_data(&netdev_shash, name);
     if (netdev) {
-        netdev_ref(netdev);
+        netdev->ref_cnt++;
     }
     ovs_mutex_unlock(&netdev_mutex);
 
@@ -2097,7 +2044,7 @@ netdev_get_devices(const struct netdev_class *netdev_class,
         struct netdev *dev = node->data;
 
         if (dev->netdev_class == netdev_class) {
-            netdev_ref(dev);
+            dev->ref_cnt++;
             shash_add(device_list, node->name, node->data);
         }
     }
@@ -2128,7 +2075,7 @@ netdev_get_vports(size_t *size)
         struct netdev *dev = node->data;
 
         if (netdev_vport_is_vport_class(dev->netdev_class)) {
-            netdev_ref(dev);
+            dev->ref_cnt++;
             vports[n] = dev;
             n++;
         }
@@ -2156,7 +2103,7 @@ netdev_get_type_from_name(const char *name)
 struct netdev *
 netdev_rxq_get_netdev(const struct netdev_rxq *rx)
 {
-    ovs_assert(ovs_refcount_read(&rx->netdev->refcount) > 0);
+    ovs_assert(rx->netdev->ref_cnt > 0);
     return rx->netdev;
 }
 
@@ -2343,10 +2290,9 @@ netdev_reconfigure(struct netdev *netdev)
 
     netdev->last_reconfigure_seq = seq_read(netdev->reconfigure_seq);
 
-    netdev->reconfigure_status = class->reconfigure
-        ? class->reconfigure(netdev)
-        : EOPNOTSUPP;
-    return netdev->reconfigure_status;
+    return (class->reconfigure
+            ? class->reconfigure(netdev)
+            : EOPNOTSUPP);
 }
 
 void
