@@ -19,6 +19,7 @@
 #include <errno.h>
 #include <inttypes.h>
 #include <stdlib.h>
+#include <sys/socket.h>
 #include <sys/types.h>
 #include <sys/uio.h>
 #include <unistd.h>
@@ -28,10 +29,11 @@
 #include "openvswitch/hmap.h"
 #include "netlink.h"
 #include "netlink-protocol.h"
+#include "netnsid.h"
 #include "odp-netlink.h"
 #include "openvswitch/ofpbuf.h"
 #include "ovs-thread.h"
-#include "poll-loop.h"
+#include "openvswitch/poll-loop.h"
 #include "seq.h"
 #include "socket-util.h"
 #include "util.h"
@@ -117,6 +119,7 @@ nl_sock_create(int protocol, struct nl_sock **sockp)
     struct nl_sock *sock;
 #ifndef _WIN32
     struct sockaddr_nl local, remote;
+    int one = 1;
 #endif
     socklen_t local_size;
     int rcvbuf;
@@ -175,7 +178,7 @@ nl_sock_create(int protocol, struct nl_sock **sockp)
     sock->protocol = protocol;
     sock->next_seq = 1;
 
-    rcvbuf = 1024 * 1024;
+    rcvbuf = 1024 * 1024 * 4;
 #ifdef _WIN32
     sock->rcvbuf = rcvbuf;
     retval = get_sock_pid_from_kernel(sock);
@@ -187,6 +190,11 @@ nl_sock_create(int protocol, struct nl_sock **sockp)
         goto error;
     }
 #else
+    if (setsockopt(sock->fd, SOL_NETLINK, NETLINK_EXT_ACK, &one, sizeof one)) {
+        VLOG_WARN_RL(&rl, "setting extended ack support failed (%s)",
+                     ovs_strerror(errno));
+    }
+
     if (setsockopt(sock->fd, SOL_SOCKET, SO_RCVBUFFORCE,
                    &rcvbuf, sizeof rcvbuf)) {
         /* Only root can use SO_RCVBUFFORCE.  Everyone else gets EPERM.
@@ -197,6 +205,15 @@ nl_sock_create(int protocol, struct nl_sock **sockp)
         }
     }
 
+    /* Strict checking only supported for NETLINK_ROUTE. */
+    if (protocol == NETLINK_ROUTE
+        && setsockopt(sock->fd, SOL_NETLINK, NETLINK_GET_STRICT_CHK,
+                      &one, sizeof one) < 0) {
+        VLOG_RL(&rl, errno == ENOPROTOOPT ? VLL_DBG : VLL_WARN,
+                "netlink: could not enable strict checking (%s)",
+                ovs_strerror(errno));
+    }
+
     retval = get_socket_rcvbuf(sock->fd);
     if (retval < 0) {
         retval = -retval;
@@ -440,6 +457,33 @@ nl_sock_join_mcgroup(struct nl_sock *sock, unsigned int multicast_group)
     return 0;
 }
 
+/* When 'enable' is true, it tries to enable 'sock' to receive netlink
+ * notifications form all network namespaces that have an nsid assigned
+ * into the network namespace where the socket has been opened. The
+ * running kernel needs to provide support for that. When 'enable' is
+ * false, it will receive netlink notifications only from the network
+ * namespace where the socket has been opened.
+ *
+ * Returns 0 if successful, otherwise a positive errno.  */
+int
+nl_sock_listen_all_nsid(struct nl_sock *sock, bool enable)
+{
+    int error;
+    int val = enable ? 1 : 0;
+
+#ifndef _WIN32
+    if (setsockopt(sock->fd, SOL_NETLINK, NETLINK_LISTEN_ALL_NSID, &val,
+                   sizeof val) < 0) {
+        error = errno;
+        VLOG_INFO("netlink: could not %s listening to all nsid (%s)",
+                  enable ? "enable" : "disable", ovs_strerror(error));
+        return errno;
+    }
+#endif
+
+    return 0;
+}
+
 #ifdef _WIN32
 int
 nl_sock_subscribe_packet__(struct nl_sock *sock, bool subscribe)
@@ -607,7 +651,7 @@ nl_sock_send_seq(struct nl_sock *sock, const struct ofpbuf *msg,
 }
 
 static int
-nl_sock_recv__(struct nl_sock *sock, struct ofpbuf *buf, bool wait)
+nl_sock_recv__(struct nl_sock *sock, struct ofpbuf *buf, int *nsid, bool wait)
 {
     /* We can't accurately predict the size of the data to be received.  The
      * caller is supposed to have allocated enough space in 'buf' to handle the
@@ -618,7 +662,10 @@ nl_sock_recv__(struct nl_sock *sock, struct ofpbuf *buf, bool wait)
     uint8_t tail[65536];
     struct iovec iov[2];
     struct msghdr msg;
+    uint8_t msgctrl[64];
+    struct cmsghdr *cmsg;
     ssize_t retval;
+    int *ptr;
     int error;
 
     ovs_assert(buf->allocated >= sizeof *nlmsghdr);
@@ -632,6 +679,8 @@ nl_sock_recv__(struct nl_sock *sock, struct ofpbuf *buf, bool wait)
     memset(&msg, 0, sizeof msg);
     msg.msg_iov = iov;
     msg.msg_iovlen = 2;
+    msg.msg_control = msgctrl;
+    msg.msg_controllen = sizeof msgctrl;
 
     /* Receive a Netlink message from the kernel.
      *
@@ -706,6 +755,41 @@ nl_sock_recv__(struct nl_sock *sock, struct ofpbuf *buf, bool wait)
     }
 #endif
 
+    if (nsid) {
+        /* The network namespace id from which the message was sent comes
+         * as ancillary data. For older kernels, this data is either not
+         * available or it might be -1, so it falls back to local network
+         * namespace (no id). Latest kernels return a valid ID only if
+         * available or nothing. */
+        netnsid_set_local(nsid);
+#ifndef _WIN32
+        cmsg = CMSG_FIRSTHDR(&msg);
+        while (cmsg != NULL) {
+            if (cmsg->cmsg_level == SOL_NETLINK
+                && cmsg->cmsg_type == NETLINK_LISTEN_ALL_NSID) {
+                ptr = ALIGNED_CAST(int *, CMSG_DATA(cmsg));
+                netnsid_set(nsid, *ptr);
+            }
+            if (cmsg->cmsg_level == SOL_SOCKET
+                && cmsg->cmsg_type == SCM_RIGHTS) {
+                /* This is unexpected and unwanted, close all fds */
+                int nfds;
+                int i;
+                nfds = (cmsg->cmsg_len - CMSG_ALIGN(sizeof(struct cmsghdr)))
+                       / sizeof(int);
+                ptr = ALIGNED_CAST(int *, CMSG_DATA(cmsg));
+                for (i = 0; i < nfds; i++) {
+                    VLOG_ERR_RL(&rl, "closing unexpected received fd (%d).",
+                                ptr[i]);
+                    close(ptr[i]);
+                }
+            }
+
+            cmsg = CMSG_NXTHDR(&msg, cmsg);
+        }
+#endif
+    }
+
     log_nlmsg(__func__, 0, buf->data, buf->size, sock->protocol);
     COVERAGE_INC(netlink_received);
 
@@ -714,7 +798,8 @@ nl_sock_recv__(struct nl_sock *sock, struct ofpbuf *buf, bool wait)
 
 /* Tries to receive a Netlink message from the kernel on 'sock' into 'buf'.  If
  * 'wait' is true, waits for a message to be ready.  Otherwise, fails with
- * EAGAIN if the 'sock' receive buffer is empty.
+ * EAGAIN if the 'sock' receive buffer is empty.  If 'nsid' is provided, the
+ * network namespace id from which the message was sent will be provided.
  *
  * The caller must have initialized 'buf' with an allocation of at least
  * NLMSG_HDRLEN bytes.  For best performance, the caller should allocate enough
@@ -730,9 +815,9 @@ nl_sock_recv__(struct nl_sock *sock, struct ofpbuf *buf, bool wait)
  * Regardless of success or failure, this function resets 'buf''s headroom to
  * 0. */
 int
-nl_sock_recv(struct nl_sock *sock, struct ofpbuf *buf, bool wait)
+nl_sock_recv(struct nl_sock *sock, struct ofpbuf *buf, int *nsid, bool wait)
 {
-    return nl_sock_recv__(sock, buf, wait);
+    return nl_sock_recv__(sock, buf, nsid, wait);
 }
 
 static void
@@ -821,7 +906,7 @@ nl_sock_transact_multiple__(struct nl_sock *sock,
         }
 
         /* Receive a reply. */
-        error = nl_sock_recv__(sock, buf_txn->reply, false);
+        error = nl_sock_recv__(sock, buf_txn->reply, NULL, false);
         if (error) {
             if (error == EAGAIN) {
                 nl_sock_record_errors__(transactions, n, 0);
@@ -840,15 +925,17 @@ nl_sock_transact_multiple__(struct nl_sock *sock,
         i = seq - base_seq;
         txn = transactions[i];
 
+        const char *err_msg = NULL;
         /* Fill in the results for 'txn'. */
-        if (nl_msg_nlmsgerr(buf_txn->reply, &txn->error)) {
+        if (nl_msg_nlmsgerr(buf_txn->reply, &txn->error, &err_msg)) {
+            if (txn->error) {
+                VLOG_DBG_RL(&rl, "received NAK error=%d - %s",
+                            txn->error,
+                            err_msg ? err_msg : ovs_strerror(txn->error));
+            }
             if (txn->reply) {
                 ofpbuf_clear(txn->reply);
             }
-            if (txn->error) {
-                VLOG_DBG_RL(&rl, "received NAK error=%d (%s)",
-                            error, ovs_strerror(txn->error));
-            }
         } else {
             txn->error = 0;
             if (txn->reply && txn != buf_txn) {
@@ -929,7 +1016,7 @@ nl_sock_transact_multiple__(struct nl_sock *sock,
             /* Handle errors embedded within the netlink message. */
             ofpbuf_use_stub(&tmp_reply, reply_buf, sizeof reply_buf);
             tmp_reply.size = sizeof reply_buf;
-            if (nl_msg_nlmsgerr(&tmp_reply, &txn->error)) {
+            if (nl_msg_nlmsgerr(&tmp_reply, &txn->error, NULL)) {
                 if (txn->reply) {
                     ofpbuf_clear(txn->reply);
                 }
@@ -1101,7 +1188,7 @@ nl_dump_refill(struct nl_dump *dump, struct ofpbuf *buffer)
     int error;
 
     while (!buffer->size) {
-        error = nl_sock_recv__(dump->sock, buffer, false);
+        error = nl_sock_recv__(dump->sock, buffer, NULL, false);
         if (error) {
             /* The kernel never blocks providing the results of a dump, so
              * error == EAGAIN means that we've read the whole thing, and
@@ -1121,7 +1208,7 @@ nl_dump_refill(struct nl_dump *dump, struct ofpbuf *buffer)
         }
     }
 
-    if (nl_msg_nlmsgerr(buffer, &error) && error) {
+    if (nl_msg_nlmsgerr(buffer, &error, NULL) && error) {
         VLOG_INFO_RL(&rl, "netlink dump request error (%s)",
                      ovs_strerror(error));
         ofpbuf_clear(buffer);
@@ -1501,6 +1588,11 @@ do_lookup_genl_family(const char *name, struct nlattr **attrs,
         family_name = OVS_WIN_NETDEV_FAMILY;
         family_version = OVS_WIN_NETDEV_VERSION;
         family_attrmax = OVS_WIN_NETDEV_ATTR_MAX;
+    } else if (!strcmp(name, OVS_CT_LIMIT_FAMILY)) {
+        family_id = OVS_WIN_NL_CTLIMIT_FAMILY_ID;
+        family_name = OVS_CT_LIMIT_FAMILY;
+        family_version = OVS_CT_LIMIT_VERSION;
+        family_attrmax = OVS_CT_LIMIT_ATTR_MAX;
     } else {
         ofpbuf_delete(reply);
         return EINVAL;
