@@ -17,6 +17,7 @@
 #ifndef OVS_TNL_ROUTER_H
 #define OVS_TNL_ROUTER_H 1
 
+#include <sys/types.h>
 #include <netinet/in.h>
 
 #include "util.h"
@@ -30,9 +31,20 @@ bool ovs_router_lookup(uint32_t mark, const struct in6_addr *ip_dst,
                        struct in6_addr *src, struct in6_addr *gw);
 void ovs_router_init(void);
 void ovs_router_insert(uint32_t mark, const struct in6_addr *ip_dst,
-                       uint8_t plen,
+                       uint8_t plen, bool local,
                        const char output_bridge[], const struct in6_addr *gw);
+void ovs_router_force_insert(uint32_t mark, const struct in6_addr *ip_dst,
+                             uint8_t plen, bool local,
+                             const char output_bridge[],
+                             const struct in6_addr *gw);
 void ovs_router_flush(void);
+
+void ovs_router_disable_system_routing_table(void);
+
+int ovs_router_get_netdev_source_address(const struct in6_addr *ip6_dst,
+                                         const char output_bridge[],
+                                         struct in6_addr *psrc);
+
 #ifdef  __cplusplus
 }
 #endif
