@@ -21,7 +21,7 @@
 #include <signal.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include "poll-loop.h"
+#include "openvswitch/poll-loop.h"
 #include "socket-util.h"
 #include "openvswitch/type-props.h"
 #include "util.h"
