@@ -364,29 +364,17 @@ void
 tnl_wc_init(struct flow *flow, struct flow_wildcards *wc)
 {
     if (tnl_port_should_receive(flow)) {
-        if (!(flow->tunnel.flags & FLOW_TNL_F_EXPLICIT)) {
-            wc->masks.tunnel.tun_id = OVS_BE64_MAX;
-            if (flow->tunnel.ip_dst) {
-                wc->masks.tunnel.ip_src = OVS_BE32_MAX;
-                wc->masks.tunnel.ip_dst = OVS_BE32_MAX;
-            } else {
-                wc->masks.tunnel.ipv6_src = in6addr_exact;
-                wc->masks.tunnel.ipv6_dst = in6addr_exact;
-            }
+        wc->masks.tunnel.tun_id = OVS_BE64_MAX;
+        if (flow->tunnel.ip_dst) {
+            wc->masks.tunnel.ip_src = OVS_BE32_MAX;
+            wc->masks.tunnel.ip_dst = OVS_BE32_MAX;
+        } else {
+            wc->masks.tunnel.ipv6_src = in6addr_exact;
+            wc->masks.tunnel.ipv6_dst = in6addr_exact;
         }
-
-        /* OpenFlow does not support the +csum flag on tunnels and
-         * netdev offloads ignore that match. A packet in SW datapath
-         * however would generate different dp-flows depending on
-         * whether its csum is set, breaking some offload implementation.
-         * Align all layers to ignore the csum in all matches, even
-         * in SW. */
         wc->masks.tunnel.flags = (FLOW_TNL_F_DONT_FRAGMENT |
-                                  /* FLOW_TNL_F_CSUM - See comment above */
+                                  FLOW_TNL_F_CSUM |
                                   FLOW_TNL_F_KEY);
-        if (flow->tunnel.flags & FLOW_TNL_F_EXPLICIT) {
-            wc->masks.tunnel.flags |= FLOW_TNL_F_EXPLICIT;
-        }
         wc->masks.tunnel.ip_tos = UINT8_MAX;
         wc->masks.tunnel.ip_ttl = 0;
         /* The tp_src and tp_dst members in flow_tnl are set to be always
@@ -428,7 +416,7 @@ tnl_port_send(const struct ofport_dpif *ofport, struct flow *flow,
         pre_flow_str = flow_to_string(flow, NULL);
     }
 
-    if (!cfg->is_explicit && !cfg->ip_src_flow) {
+    if (!cfg->ip_src_flow) {
         flow->tunnel.ip_src = in6_addr_get_mapped_ipv4(&tnl_port->match.ipv6_src);
         if (!flow->tunnel.ip_src) {
             flow->tunnel.ipv6_src = tnl_port->match.ipv6_src;
@@ -436,7 +424,7 @@ tnl_port_send(const struct ofport_dpif *ofport, struct flow *flow,
             flow->tunnel.ipv6_src = in6addr_any;
         }
     }
-    if (!cfg->is_explicit && !cfg->ip_dst_flow) {
+    if (!cfg->ip_dst_flow) {
         flow->tunnel.ip_dst = in6_addr_get_mapped_ipv4(&tnl_port->match.ipv6_dst);
         if (!flow->tunnel.ip_dst) {
             flow->tunnel.ipv6_dst = tnl_port->match.ipv6_dst;
@@ -444,11 +432,10 @@ tnl_port_send(const struct ofport_dpif *ofport, struct flow *flow,
             flow->tunnel.ipv6_dst = in6addr_any;
         }
     }
-    if (!cfg->is_explicit) {
-        flow->tunnel.tp_dst = cfg->dst_port;
-        if (!cfg->out_key_flow) {
-            flow->tunnel.tun_id = cfg->out_key;
-        }
+    flow->tunnel.tp_src = 0; /* Do not carry from a previous tunnel. */
+    flow->tunnel.tp_dst = cfg->dst_port;
+    if (!cfg->out_key_flow) {
+        flow->tunnel.tun_id = cfg->out_key;
     }
 
     if (cfg->ttl_inherit && is_ip_any(flow)) {
@@ -479,7 +466,7 @@ tnl_port_send(const struct ofport_dpif *ofport, struct flow *flow,
     flow->tunnel.flags &= ~(FLOW_TNL_F_MASK & ~FLOW_TNL_PUB_F_MASK);
     flow->tunnel.flags |= (cfg->dont_fragment ? FLOW_TNL_F_DONT_FRAGMENT : 0)
         | (cfg->csum ? FLOW_TNL_F_CSUM : 0)
-        | ((cfg->out_key_present | cfg->is_explicit) ? FLOW_TNL_F_KEY : 0);
+        | (cfg->out_key_present ? FLOW_TNL_F_KEY : 0);
 
     if (cfg->set_egress_pkt_mark) {
         flow->pkt_mark = cfg->egress_pkt_mark;
