@@ -7,7 +7,7 @@
 #include <inttypes.h>
 #include <stdarg.h>
 #include <stdlib.h>
-#include "ofpbuf.h"
+#include "openvswitch/ofpbuf.h"
 #include "ovstest.h"
 #include "dp-packet.h"
 #include "packets.h"
@@ -466,8 +466,10 @@ test_rstp_main(int argc, char *argv[])
     vlog_set_pattern(VLF_CONSOLE, "%c|%p|%m");
     vlog_set_levels(NULL, VLF_SYSLOG, VLL_OFF);
 
+    rstp_init();
+
     if (argc != 2) {
-        ovs_fatal(0, "usage: test-rstp INPUT.RSTP\n");
+        ovs_fatal(0, "usage: test-rstp INPUT.RSTP");
     }
     file_name = argv[1];
 
@@ -691,6 +693,7 @@ test_rstp_main(int argc, char *argv[])
         }
     }
     free(token);
+    fclose(input_file);
 
     for (i = 0; i < tc->n_lans; i++) {
         struct lan *lan = tc->lans[i];
